%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.1, (the "License"); you may not use this file except in
%% compliance with the License. You should have received a copy of the
%% Erlang Public License along with this software. If not, it can be
%% retrieved via the world wide web at http://www.erlang.org/.
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Initial Developer of the Original Code is Mobile Arts AB
%% Portions created by Mobile Arts are Copyright 2002, Mobile Arts AB
%% All Rights Reserved.''
%% 
%%

%%% This version of the HTTP/1.1 client implements:
%%%      - RFC 2616 HTTP 1.1 client part
%%%      - RFC 2817 Upgrading to TLS Within HTTP/1.1 (not yet!)
%%%      - RFC 2818 HTTP Over TLS
%%%      - RFC 2965 HTTP State Management Mechanism (Still Beta)
%%%      - RFC 3229 Delta encoding in HTTP (not yet!)
%%%      - RFC 3230 Instance Digests in HTTP (not yet!)
%%%      - RFC 3310 Authentication and Key Agreement (AKA) (not yet!)
%%%      - Netscape style Cookies, see
%%%            http://wp.netscape.com/newsref/std/cookie_spec.html (Still Beta)
%%%      - HTTP/1.1 Specification Errata found at
%%%        http://skrb.org/ietf/http_errata.html
%%%    Additionaly follows the following recommendations:
%%%      - RFC 2964 Use of HTTP State Management (not yet!)
%%%      - RFC 3143 Known HTTP Proxy/Caching Problems (not yet!)
%%%      - draft-nottingham-hdrreg-http-00.txt (not yet!)
%%%
%%% Depends on
%%%      - uri.erl for all URL parsing (except what is handled by the C driver)
%%%      - http_lib.erl for all parsing of body and headers
%%%
%%% Supported Settings are:
%%% http_timeout      % (int) Milliseconds before a request times out
%%% http_useproxy     % (bool) True if a proxy should be used
%%% http_proxy        % (string) Proxy
%%% http_noproxylist  % (list) List with hosts not requiring proxy
%%% http_autoredirect % (bool) True if automatic redirection on 30X responses
%%% http_ssl          % (list) SSL settings. A non-empty list enables SSL/TLS
%%%                      support in the HTTP client
%%% http_pipelinesize % (int) Length of pipeline. 1 means no pipeline.
%%%                      Only has effect when initiating a new session.
%%% http_sessions     % (int) Max number of open sessions for {Addr,Port}
%%% http_cookie       % (bool) True if cookie headers should be handled
%%% http_relaxed      % (bool) True if behaviour should be not strictly RFC 2616
%%%                            compliant.
%%%
%%% TODO: (Known bugs!)
%% - Cache handling
%% - Doesn't handle a bunch of entity headers properly
%% - Better handling of status codes different from 200,30X and 50X 
%% - close_session/2 and cancel_request/1 doesn't work
%% - Variable pipe size.
%% - Due to the fact that inet_drv only has a single timer, the timeouts given
%%   for pipelined requests are not ok (too long)
%%
%% Note:
%% - Some servers (e.g. Microsoft-IIS/5.0) may sometimes not return a proper
%%   'Location' header on a redirect.
%%   The client will normally fail with {error,no_scheme} in these cases.
%%   Setting the http_relaxed option will cause the client to interpret the
%%   value as relativeURI 
-module(http).
-author("johan.blom@mobilearts.se").

-export([start/0,
	 request/3,request/4,cancel_request/1,
	 request_sync/2,request_sync/3]).

-include("http.hrl").
-include("jnets_httpd.hrl").

-define(START_OPTIONS,[]).

%%% HTTP Client manager. Used to store open connections.
%%% Will be started automatically unless started explicitly.
start() ->
    application:start(ssl),
    httpc_manager:start().

%%% Asynchronous HTTP request that spawns a handler.
%%% Method                          HTTPReq
%%% options,get,head,delete,trace = {Url,Headers}
%%% post,put                      = {Url,Headers,ContentType,Body}
%%%  where Url is a {Scheme,Host,Port,Path,Query} tuple, as returned by uri.erl
%%% 
%%% Returns: {ok,ReqId} |
%%%          {error,Reason}
%%% If {ok,ReqId} was returned, the handler will return with
%%%    gen_server:cast(From,{Ref,ReqId,{error,Reason}}) |
%%%    gen_server:cast(From,{Ref,ReqId,{Status,Headers,Body}})
%%%  where Reason is an atom and Headers a #http_response_h{} record
%%% http:format_error(Reason) gives a more informative description.
%%% 
%%% Note:
%%% - Always try to find an open connection to a given host and port, and use
%%%   the associated socket.
%%% - Unless a 'Connection: close' header is provided don't close the socket
%%%   after a response is given 
%%% - A given Pid, found in the database, might be terminated before the
%%%   message is sent to the Pid. This will happen e.g., if the connection is
%%%   closed by the other party and there are no pending requests.
%%% - The HTTP connection process is spawned, if necessary, in
%%%   httpc_manager:add_connection/4
request(Ref,Method,HTTPReqCont) ->
    request(Ref,Method,HTTPReqCont,[],self()).

request(Ref,Method,HTTPReqCont,Settings) ->
    request(Ref,Method,HTTPReqCont,Settings,self()).

request(Ref,Method,{{Scheme,Host,Port,Path,Query},
		    Headers,ContentType,Body},Settings,From) ->

    case create_settings(Settings,#client_settings{}) of
	{error,Reason} ->
	    {error,Reason};
	CS ->
	    case create_headers(Headers,#http_request_h{},Host) of
		{error,Reason} ->
		    {error,Reason};
		H ->
		    Req=#request{ref=Ref,from=From,
				 scheme=Scheme,address={Host,Port},
				 path=Path,pquery=Query,method=Method,
				 headers=H,content={ContentType,Body},
				 settings=CS},
		    httpc_manager:request(Req)
	    end
    end;
request(Ref,Method,{Url,Headers},Settings,From) ->
    request(Ref,Method,{Url,Headers,[],[]},Settings,From).

%%% Cancels requests identified with ReqId.
%%% FIXME! Doesn't work... 
cancel_request(ReqId) ->
    httpc_manager:cancel_request(ReqId).

%%% Close all sessions currently open to Host:Port
%%% FIXME! Doesn't work... 
close_session(Host,Port) ->
    httpc_manager:close_session(Host,Port).
    

%%% Synchronous HTTP request that waits until a response is created
%%% (e.g. successfull reply or timeout)
%%% Method                          HTTPReq
%%% options,get,head,delete,trace = {Url,Headers}
%%% post,put                      = {Url,Headers,ContentType,Body}
%%%  where Url is a string() or a {Scheme,Host,Port,Path,Query} tuple
%%% 
%%% Returns: {Status,Headers,Body} |
%%%          {error,Reason}
%%% where Reason is an atom. 
%%% http:format_error(Reason) gives a more informative description.
request_sync(Method,HTTPReqCont) ->
    request_sync(Method,HTTPReqCont,[]).

request_sync(Method,{Url,Headers},Settings)
  when Method==options;Method==get;Method==head;Method==delete;Method==trace ->
    case uri:parse(Url) of
	{error,Reason} ->
	    {error,Reason};
	ParsedUrl ->
	    request_sync2(Method,{ParsedUrl,Headers,[],[]},Settings)
    end;
request_sync(Method,{Url,Headers,ContentType,Body},Settings)
  when Method==post;Method==put ->
    case uri:parse(Url) of
	{error,Reason} ->
	    {error,Reason};
	ParsedUrl ->
	    request_sync2(Method,{ParsedUrl,Headers,ContentType,Body},Settings)
    end;
request_sync(Method,Request,Settings) ->
    {error,bad_request}.

request_sync2(Method,HTTPCont,Settings) ->
    case request(request_sync,Method,HTTPCont,Settings,self()) of
	{ok,_ReqId} ->
	    receive
		{'$gen_cast',{request_sync,_ReqId2,{Status,Headers,Body}}} ->
		    {Status, key_value_tuples(Headers),binary_to_list(Body)};
		{'$gen_cast',{request_sync,_ReqId2,{error,Reason}}} ->
		    {error,Reason}
	    end;
	Error ->
	    Error
    end.


create_settings([],Out) ->
    Out;
create_settings([{http_timeout,Val}|Settings],Out) ->
    create_settings(Settings,Out#client_settings{timeout=Val});
create_settings([{http_useproxy,Val}|Settings],Out) ->
    create_settings(Settings,Out#client_settings{useproxy=Val});
create_settings([{http_proxy,Val}|Settings],Out) ->
    create_settings(Settings,Out#client_settings{proxy=Val});
create_settings([{http_noproxylist,Val}|Settings],Out) ->
    create_settings(Settings,Out#client_settings{noproxylist=Val});
create_settings([{http_autoredirect,Val}|Settings],Out) ->
    create_settings(Settings,Out#client_settings{autoredirect=Val});
create_settings([{http_ssl,Val}|Settings],Out) ->
    create_settings(Settings,Out#client_settings{ssl=Val});
create_settings([{http_pipelinesize,Val}|Settings],Out)
  when integer(Val),Val>0 ->
    create_settings(Settings,Out#client_settings{max_quelength=Val});
create_settings([{http_sessions,Val}|Settings],Out)
  when integer(Val),Val>0 ->
    create_settings(Settings,Out#client_settings{max_sessions=Val});
create_settings([{http_relaxed,Val}|Settings],Out) when Val==true;Val==false ->
    create_settings(Settings,Out#client_settings{relaxed=Val});
create_settings([{http_cookie,Val}|Settings],Out) when Val==true;Val==false ->
    create_settings(Settings,Out#client_settings{cookie=Val});
create_settings([{Key,_Val}|_Settings],_Out) ->
    io:format("ERROR bad settings, got ~p~n",[Key]),
    {error,bad_settings}.


%%% Create a #http_request_h{} record with all HTTP headers.  Note: -
%%% Headers accepted both as a key/val list and as a #http_request_h{}
%%% record directly. In both cases a few sanity checks are necessary.
%%% %% These are done on the resulting record - No problem if headers
%%% with a dedicated field in #http_request_h{} is added to other,
%%% unless we are actually interested to check its value somewhere.  -
create_headers(Req,_,Host) when record(Req,http_request_h) ->
    validate_headers(Req,Host);
create_headers([],Req,Host) ->
    validate_headers(Req,Host);
create_headers([{Key,Val}|Rest],Req,Host) ->
    case httpd_util:to_lower(Key) of
	"expect" ->
	    create_headers(Rest,Req#http_request_h{expect=Val},Host);
	"host" ->
	    create_headers(Rest,Req#http_request_h{host=Val},Host);
	"te" ->
	    create_headers(Rest,Req#http_request_h{te=Val},Host);
	OtherKey ->
	    create_headers(Rest,
			   Req#http_request_h{
			     other=[{OtherKey,Val}|
				    Req#http_request_h.other]},Host)
    end.


validate_headers(Req=#http_request_h{te=undefined},Host) ->
    validate_headers(Req#http_request_h{te=""},Host);
validate_headers(Req=#http_request_h{host=undefined},Host) ->
    validate_headers(Req#http_request_h{host=Host},Host);
validate_headers(Req,_) ->
    Req.

key_value_tuples(Headers) ->
    HeaderFields = record_info(fields, http_response_h) -- [other],
    HeaderList = lists:foldl(fun(Key, Acc) -> 
				     case key_value_tuple(Key, Headers) of
					 undefined ->
					     Acc;
					 Tuple ->
					     [Tuple | Acc]
				     end
			     end,
			     [], HeaderFields),
    lists:reverse(HeaderList) ++ Headers#http_response_h.other.


key_value_tuple(connection, Headers) ->
    key_value_tuple("Connection", Headers#http_response_h.connection);
key_value_tuple(transfer_encoding, Headers) ->
    key_value_tuple("Transfer-Encoding", 
		    Headers#http_response_h.transfer_encoding);
key_value_tuple(retry_after, Headers) ->
    key_value_tuple("Retry-After", Headers#http_response_h.retry_after);
key_value_tuple(content_length, Headers) ->
    key_value_tuple("Content-Length", Headers#http_response_h.content_length);
key_value_tuple(content_type, Headers) ->
    key_value_tuple("Content-Type", Headers#http_response_h.content_type);
key_value_tuple(location, Headers) ->
    key_value_tuple("Location", Headers#http_response_h.location);

key_value_tuple(_, undefined) ->
    undefined;
key_value_tuple("Content-Length", "0") ->
    undefined;
key_value_tuple(Key, Value) ->
    {Key, Value}.

