%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.1, (the "License"); you may not use this file except in
%% compliance with the License. You should have received a copy of the
%% Erlang Public License along with this software. If not, it can be
%% retrieved via the world wide web at http://www.erlang.org/.
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Initial Developer of the Original Code is Mobile Arts AB
%% Portions created by Mobile Arts are Copyright 2002, Mobile Arts AB
%% All Rights Reserved.''
%% 
%%
-module(http_chunk).

-include("http.hrl").

-export([decode/2]).

decode(ResponseHeaderRcord, ChunkedBody) -> 
    {Body, Headers, Length} = decode_body(ChunkedBody, <<>>, 0),
    NewHeaders = http_response:headers(Headers, ResponseHeaderRcord),
    TransferEncoding = 
	case NewHeaders#http_response_h.transfer_encoding -- "chunked" of
	    ""  ->
		undefined;
	    Other ->
		Other
	    end,
    {NewHeaders#http_response_h{content_length = Length,
				transfer_encoding = TransferEncoding}, Body}.

decode_body(ChunkedBody, Body, AccLength) ->
    {Size, Chunk} = decode_size(ChunkedBody, []),
    {Data, ChunkRest} = decode_data(Size, Chunk),
    case is_last_chunk(ChunkRest) of
	true ->
	    <<"0", Rest/binary>> = ChunkRest,
	    <<?CR, ?LF, Trailers>> = ignore_extensions(Rest),
	    {<<Body/binary, Data/binary>>, 
	     decode_trailer(Trailers, [], []), AccLength};
	false ->
	    decode_body(ChunkRest, <<Body/binary, Data/binary>>, 
			AccLength + Size)
    end.

decode_size(<<?CR, ?LF, ChunkRest/binary>>, HexList) ->
    ChunkSize =  httpd_util:hexlist_to_integer(lists:reverse(HexList)),
    {ChunkSize, ChunkRest};

decode_size(<<";", Rest/binary>>, HexList) ->
    ChunkRest = ignore_extensions(Rest),
    decode_size(ChunkRest, HexList);

decode_size(<<Octet, Rest/binary>>, HexList) ->
    decode_size(Rest, [Octet | HexList]).

%% "All applications MUST ignore chunk-extension extensions they
%% do not understand.", see RFC 2616 Section 3.6.1 We don't
%% understand any extension...
ignore_extensions(ChunkRest = <<?CR, _>>) ->
    ChunkRest;
ignore_extensions(<<Octet, Rest/binary>>) ->
    ignore_extensions(Rest).

decode_data(ChunkSize, TotalChunk) ->
    <<Data:ChunkSize/binary, Rest/binary>> = TotalChunk,
    {Data, Rest}.  

decode_trailer(<<?CR,?LF>>, Header, Headers) ->
    [lists:reverse(Header) | Headers];

decode_trailer(<<?CR, ?LF, Rest/binary>>, Header, Headers) ->
    decode_trailer(Rest, [], [lists:reverse(Header) | Headers]);

decode_trailer(<<A, B, Rest/binary>>, Header, Headers) ->
  decode_trailer(Rest, [B, A | Header], Headers).

is_last_chunk(<<"0", Rest/binary>>) ->
    true;
is_last_chunk(_) ->
    false.


