%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.1, (the "License"); you may not use this file except in
%% compliance with the License. You should have received a copy of the
%% Erlang Public License along with this software. If not, it can be
%% retrieved via the world wide web at http://www.erlang.org/.
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Initial Developer of the Original Code is Mobile Arts AB
%% Portions created by Mobile Arts are Copyright 2002, Mobile Arts AB
%% All Rights Reserved.''
%% 
%%
-module(http_response).

-include("http.hrl").

-export([parse_message/1, parse_body/2, headers/2, respond/3, error/2]).

%%%=========================================================================
%%%  API
%%%=========================================================================

%%-------------------------------------------------------------------------
%% parse_message(Packet) -> {StatusCode, Headers, Body}
%%	Packet - binary()
%%	StatusCode - integer()
%%	Headers - #http_response_h{}
%%                                   
%% Description: Parses a http_respons message
%%-------------------------------------------------------------------------
parse_message(Packet) ->
    {StatusCode, Rest} = parse_status_line(Packet),
    {Headers, Body} = parse_headers(Rest, [], []),
    {StatusCode, Headers, Body}.

%%-------------------------------------------------------------------------
%% parse_body(Headers, Body) ->
%%	Headers - #http_response_h{}
%%	Body - binary()
%%                                   
%% Description: Decodes body if necessary. 
%%-------------------------------------------------------------------------
parse_body(Headers, Body) ->
    case Headers#http_response_h.transfer_encoding of
	"chunked" ->
	    http_chunk:decode(Body, Headers);
	Encoding when list(Encoding) ->
	    throw({error, unknown_coding});
	_ ->
	    {Headers, Body}
    end.

%%-------------------------------------------------------------------------
%% headers(HeaderList, #http_response_h{}) -> #http_response_h{}
%%   HeaderList - ["HeaderField:Value"]     	
%%   HeaderField - string()
%%   Value - string()	
%%                                   
%% Description: Creates a http_response_h-record used internally to
%%              handle http-headers.
%%-------------------------------------------------------------------------
headers([], Headers) ->
    Headers;

headers([Header | Tail], Headers) ->  
    {Key, [$: | Value]} =
	lists:splitwith(fun($:) -> false; (_) -> true end, Header), 
    headers(Tail, headers(http_lib:strip(Key), 
			  http_lib:strip(Value), Headers)).

%%-------------------------------------------------------------------------
%% respond(Response, Request, Session) ->
%%   Response -
%%   Request - 
%%   Session -
%%                                   
%% Description: Checks the status code ...
%%-------------------------------------------------------------------------
respond(Response = {100, _, _}, Request, Session) ->
    status_continue(Request, Session);

respond({101, Headers, Body}, Request, Session) ->
    ok;
respond(Response = {200, Headers, _}, Request, Session) ->
    gen_server:cast(Request#request.from,
		    {Request#request.ref, Request#request.id,
		     Response}),
    ServerClose = http_lib:connection_close(Headers),
    handle_connection(Session#tcp_session.clientclose,ServerClose,
		      Request, Session);
respond({201, Headers, Body}, Request, Session) ->
    ok;
respond({202, Headers, Body}, Request, Session) ->
    ok;
respond({203, Headers, Body}, Request, Session) ->
    ok;
respond({204, Headers, Body}, Request, Session) ->
    ok;
respond({205, Headers, Body}, Request, Session) ->
    ok;
respond({206, Headers, Body}, Request, Session) ->
    ok;

respond(Response = {300, _, _}, Request, Session) ->
    status_multiple_choices(Response, Request, Session);
respond(Response = {301, _, _}, Request, Session) ->
    status_moved_permanently(Response, Request,Session);
respond(Response = {302, _, _}, Request, Session) ->
    status_found(Response, Request, Session);
respond(Response = {303, _, _}, Request, Session) ->
    status_see_other(Response, Request, Session);
respond(Response = {304, _, _}, Request, Session) ->
    status_not_modified(Response, Request,Session);

respond(Response = {305, _, _}, Request, Session) ->
    status_use_proxy(Response, Request, Session);
respond(Response = {307, _, _}, Request, Session) ->
    status_temporary_redirect(Response, Request,Session);
respond({400, Headers, Body}, Request, Session) ->
    ok;
respond({401, Headers, Body}, Request, Session) ->
    ok;
respond({402, Headers, Body}, Request, Session) ->
    ok;
respond({403, Headers, Body}, Request, Session) ->
    ok;
respond({404, Headers, Body}, Request, Session) ->
    ok;
respond({405, Headers, Body}, Request, Session) ->
    ok;
respond({406, Headers, Body}, Request, Session) ->
    ok;
respond({407, Headers, Body}, Request, Session) ->
    ok;
respond({408, Headers, Body}, Request, Session) ->
    ok;
respond({409, Headers, Body}, Request, Session) ->
    ok;
respond({410, Headers, Body}, Request, Session) ->
    ok;
respond({411, Headers, Body}, Request, Session) ->
    ok;
respond({412, Headers, Body}, Request, Session) ->
    ok;
respond({413, Headers, Body}, Request, Session) ->
    ok;
respond({414, Headers, Body}, Request, Session) ->
    ok;
respond({415, Headers, Body}, Request, Session) ->
    ok;
respond({416, Headers, Body}, Request, Session) ->
    ok;
respond({417, Headers, Body}, Request, Session) ->
    ok;

respond(Response = {500, _, _}, Request, Session) ->
     status_server_error_50x(Response, Request, Session);

respond(Response = {501, _, _}, Request, Session) ->
     status_server_error_50x(Response, Request, Session);

respond(Response = {502, _, _}, Request, Session) ->
     status_server_error_50x(Response, Request, Session);

respond(Response = {503, _, _}, Request, Session) ->
    status_service_unavailable(Response, Request, Session);

respond(Response = {504, _, _}, Request, Session) ->
    status_server_error_50x(Response, Request, Session);

respond(Response = {505, _, _}, Request, Session) ->
    status_server_error_50x(Response, Request, Session).

%%%========================================================================
%%% Internal functions
%%%========================================================================
parse_status_line(Packet) ->
    {_, RestPacket} = parse_version(Packet, []),
    {StatusCode, TmpRestPacket} = parse_status_code(RestPacket, []),
    {_, NewRestPacket} = parse_reason_phrase(TmpRestPacket, []),
    {StatusCode, NewRestPacket}.

parse_version(<<?SP, Rest/binary>>, Version) ->
    {lists:reverse(Version), Rest};
parse_version(<<Octet, Rest/binary>>, Version) ->
    parse_version(Rest, [Octet | Version]).

parse_status_code(<<?SP, Rest/binary>>, Acc) -> 
    {list_to_integer(lists:reverse(Acc)), Rest};

parse_status_code(<<Octet, Rest/binary>>, Acc) ->
    parse_status_code(Rest, [Octet | Acc]).

parse_reason_phrase(<<?CR, ?LF, Rest/binary>>, Phrase) ->
    {lists:reverse(Phrase), Rest};	 
		  
parse_reason_phrase(<<Octet, Rest/binary>>, Phrase) ->
    parse_reason_phrase(Rest, [Octet | Phrase]).

parse_headers(<<?CR,?LF,?CR,?LF,Body/binary>>, Header, Headers) ->
    ResponseHeaderRcord = headers([lists:reverse(Header) | Headers],
				  #http_response_h{}),
    {ResponseHeaderRcord, Body};

parse_headers(<<?CR,?LF, Octet, Rest/binary>>, Header, Headers) ->
    parse_headers(Rest, [Octet], [lists:reverse(Header) | Headers]);

parse_headers(<<Octet, Rest/binary>>, Header, Headers) ->
    parse_headers(Rest, [Octet | Header], Headers).

headers("Connection", Value, Headers) ->
    Headers#http_response_h{connection = Value};
headers("Content-Length", Value, Headers) ->
    Headers#http_response_h{content_length = Value};
headers("Content-Type", Value, Headers) ->
    Headers#http_response_h{content_type = Value};
headers("Transfer-Encoding", Value, Headers) ->
    Headers#http_response_h{transfer_encoding = Value};
headers("Location", Value, Headers) ->
    Headers#http_response_h{location = Value};
headers("Retry-After", Value, Headers) ->
    Headers#http_response_h{retry_after = Value};
headers(Key, Value, Headers) ->
    Headers#http_response_h{other=
			    [{Key, Value} | Headers#http_response_h.other]}.

%%% RFC2616, Section 10.1.1
%%% Note:
%%% - Only act on the 100 status if the request included the
%%%   "Expect:100-continue" header, otherwise just ignore this response.
status_continue(Req = #http_request_h{expect="100-continue"}, Session) ->
    {_, Body} = Req#request.content,
    http_lib:send(Session#tcp_session.scheme, 
		  Session#tcp_session.socket, Body),
    httpc_handler:next_response_with_request(Req, Session);

status_continue(Req, Session) ->
    httpc_handler:next_response_with_request(Req, Session).	

%%% RFC2616, Section 10.3.1: Multiple Choices 
%%% Note: - If response to a HEAD request, the Content-Type/Body both
%%% should be empty. The behaviour on an empty Content-Type or Body
%%% is unspecified.  However, e.g. "Apache/1.3" servers returns both
%%% empty if the header 'if-modified-since: Date' was sent in the
%%% request and the content is "not modified" (instead of 304). Thus
%%% implicitly giving the cache as the only choice.
status_multiple_choices(Response = {_, Headers, _}, 
			Request = #request{settings = 
					   #client_settings{autoredirect =
							    true}
					  }, Session) ->
    ServerClose = http_lib:connection_close(Headers),
    redirect(Response, Request, Session, fun status_multiple_choices/3,
	    ServerClose);

status_multiple_choices(Response = {_, Headers, _}, Req, Session) ->
    ServerClose = http_lib:connection_close(Headers),
    gen_server:cast(Req#request.from,{Req#request.ref,Req#request.id,
				      Response}),
    handle_connection(Session#tcp_session.clientclose, ServerClose, Req, 
		      Session).

%%% RFC2616, Section 10.3.2: Moved Permanently
status_moved_permanently(Response = {_, Headers, _}, 
			 Request = #request{settings = 
					    #client_settings{autoredirect 
							     = true}
					   }, Session) ->
    
    ServerClose = http_lib:connection_close(Headers),
    redirect(Response, Request, Session, fun status_moved_permanently/3,
	     ServerClose);  

status_moved_permanently(Response =  {_, Headers, _}, Request, Session) ->
    ServerClose = http_lib:connection_close(Headers),
    gen_server:cast(Request#request.from, {Request#request.ref,
					   Request#request.id,
					   Response}),
    handle_connection(Session#tcp_session.clientclose, ServerClose,
		      Request, Session).

%%% RFC2616, Section 10.3.3: Found
status_found(Response = {_, Headers, _}, 
	     Request = #request{settings = 
				#client_settings{autoredirect = true}}, 
	     Session) ->
    ServerClose = http_lib:connection_close(Headers),
    redirect(Response, Request, Session, fun status_found/3,
	     ServerClose);  

status_found(Response = {_, Headers, _}, Request, Session) ->
    ServerClose = http_lib:connection_close(Headers),
    gen_server:cast(Request#request.from,
		    {Request#request.ref,Request#request.id,
				      Response}),
    handle_connection(Session#tcp_session.clientclose, ServerClose,
		      Request, Session).

%%% RFC2616,  Section 10.3.4: See Other
status_see_other(Response = {_, Headers, _}, 
		 Request = #request{settings = 
				    #client_settings{autoredirect = true}}, 
		 Session) ->
    
    ServerClose = http_lib:connection_close(Headers),
    redirect(Response, Request, Session, fun status_see_other/3,
	    ServerClose);  

status_see_other(Response = {_, Headers, _},Request, Session) ->
    ServerClose = http_lib:connection_close(Headers),
    gen_server:cast(Request#request.from, {Request#request.ref, 
					   Request#request.id,
					   Response}),
    handle_connection(Session#tcp_session.clientclose, ServerClose,
		      Request, Session).

%%% RFC2616,  Section 10.3.5: Not Modified
status_not_modified(Response = {_, Headers, _}, 
		    Request =  #request{settings = 
					#client_settings{autoredirect 
							 = true}}, 
		    Session) ->
    
    ServerClose = http_lib:connection_close(Headers),
    redirect(Response, Request, Session, fun status_see_other/3,
	    ServerClose);  
   
status_not_modified(Response = {_, Headers, _}, Request, Session) ->
    ServerClose=http_lib:connection_close(Headers),
    gen_server:cast(Request#request.from,
		    {Request#request.ref,Request#request.id,
		     Response}),
    handle_connection(Session#tcp_session.clientclose,ServerClose,
		      Request,Session).


%%% RFC2616, Section 10.3.6: Use Proxy
status_use_proxy(Response = {_, Headers, _}, 
		 Request = #request{settings = 
				    #client_settings{autoredirect = true}}, 
		 Session) ->
    ServerClose = http_lib:connection_close(Headers),
    redirect(Response, Request, Session, fun status_use_proxy/3,
	     ServerClose);  
    
status_use_proxy(Response = {_, Headers, _},Request, Session) ->
    ServerClose = http_lib:connection_close(Headers),
    gen_server:cast(Request#request.from, 
		    {Request#request.ref,Request#request.id, Response}),
    handle_connection(Session#tcp_session.clientclose, ServerClose, 
		      Request, Session).


%%% RFC2616, Section 10.3.8: Temporary Redirect
status_temporary_redirect(Response = {_, Headers, _}, 
			  Request = #request{settings = 
					     #client_settings{autoredirect = 
							      true}}, 
			  Session) ->
    
    ServerClose = http_lib:connection_close(Headers),
    redirect(Response, Request, Session, fun status_temporary_redirect/3,
	    ServerClose);

status_temporary_redirect(Response = {_, Headers, _}, Request, Session) ->
    ServerClose = http_lib:connection_close(Headers),
    gen_server:cast(Request#request.from, 
		    {Request#request.ref,Request#request.id,
		     Response}),
    handle_connection(Session#tcp_session.clientclose, ServerClose,
		      Request, Session).

%%% RFC2616, Section 10.5.4: Service Unavailable
%%%    The server is currently unable to handle the request due to a
%%%    temporary overloading or maintenance of the server. The implication
%%%    is that this is a temporary condition which will be alleviated after
%%%    some delay. If known, the length of the delay MAY be indicated in a
%%%    Retry-After header. If no Retry-After is given, the client SHOULD
%%%    handle the response as it would for a 500 response.
%% Note:
%% - This session is now considered busy, thus cancel any requests in the
%%   pipeline and close the session.
%% FIXME! Implement a user option to automatically retry if the 'Retry-After'
%%        header is given.
status_service_unavailable(Resp,Req,Session) ->
%    RetryAfter=Headers#http_response_h.retry_after,    
    gen_server:cast(Req#request.from,{Req#request.ref,Req#request.id,Resp}),
    close_session(server_connection_close,Req,Session).    

%%% Received a 50x Status code (~ "Service Error")
%%%   Response status codes beginning with the digit "5" indicate cases in
%%%   which the server is aware that it has erred or is incapable of
%%%   performing the request.
status_server_error_50x(Resp,Req,Session) ->
    gen_server:cast(Req#request.from,{Req#request.ref,Req#request.id,Resp}),
    close_session(server_connection_close,Req,Session).    


redirect(Response = {StatusCode, Headers, Body}, Req, Session, Fun, 
	 ServerClose) ->
   
    case Headers#http_response_h.location of
	undefined ->
	    gen_server:cast(Req#request.from,{Req#request.ref,Req#request.id,
					      Response}),
	    handle_connection(Session#tcp_session.clientclose, ServerClose,
			      Req, Session);
	RedirUrl ->
	    Scheme = Session#tcp_session.scheme,
	    case uri:parse(RedirUrl) of
		{error, no_scheme} when
		      (Req#request.settings)#client_settings.relaxed ->
		    NewLocation=fix_relative_uri(Req,RedirUrl),
		    Fun({StatusCode, Headers#http_response_h{
				       location=NewLocation},
			 Body},Req,Session);
		{error, Reason} ->
		    {error,  Reason};
		{Scheme, Host, Port, Path,  Query} -> % Automatic redirection
		    NewHeaders = (Req#request.headers)#http_request_h{host 
								      = Host},
		    NewReq = Req#request{redircount = Req#request.redircount+1,
					 headers = NewHeaders,
					 address = {Host,Port},
					 path = Path,
					 pquery = Query},
		    handle_redirect(Session#tcp_session.clientclose,
				    ServerClose,
				    NewReq,Session)
	    end
    end.

%%% Guessing that we received a relative URI, fix it to become an absoluteURI
fix_relative_uri(Req,RedirUrl) ->
    {Server,Port}=Req#request.address,
    Path=Req#request.path,
    atom_to_list(Req#request.scheme)++"://"++Server++Path++RedirUrl.
    

%%% Handles requests for redirects
%%% The redirected request might be:
%%% - FIXME! on another TCP session, another scheme
%%% - on the same TCP session, same scheme
%%% - on another TCP session , same scheme
%%% However, in all cases treat it as a new request, with redircount updated.
%%%
%%% The redirect may fail, but this not a reason to close this session.
%%% Instead return a error for this request, and continue as ok.
handle_redirect(ClientClose,ServerClose,Req,Session) ->
    case httpc_manager:request(Req) of
	{ok,_ReqId} -> % FIXME Should I perhaps reuse the Reqid?
	    handle_connection(ClientClose,ServerClose,Req,Session);
	{error,Reason} ->
	    error(Req,Reason),
	    handle_connection(ClientClose,ServerClose,Req,Session)
    end.

%%% Check if the persistent connection flag is false (ie client request
%%% non-persistive connection), or if the server requires a closed connection
%%% (by sending a "Connection: close" header). If the connection required
%%% non-persistent, we may close the connection immediately.
handle_connection(ClientClose,ServerClose,Req,Session) ->
    case {ClientClose,ServerClose} of
	{false,false} ->
	    ok;
	{false,true} -> % The server requests this session to be closed.
	    close_session(server_connection_close,Req,Session);
	{true,_} -> % The client requested a non-persistent connection
	    close_session(client_connection_close,Req,Session)
    end.


%%% Close the session.
%%% We now have three cases:
%%% - Client request a non-persistent connection when initiating the request.
%%%   Session info not stored in httpc_manager
%%% - Server requests a non-persistent connection when answering a request.
%%%   No need to resend request, but there might be a pipeline.
%%% - Some kind of error
%%%   Close the session, we may then try resending all requests in the pipeline
%%%   including the current depending on the error.
%%% FIXME! Should not always abort the session (see close_session in
%%%     httpc_manager for more details)
close_session(client_connection_close,_Req,Session) ->
    http_lib:close(Session#tcp_session.scheme,Session#tcp_session.socket),
    stop;
close_session(server_connection_close,Req,Session) ->
    http_lib:close(Session#tcp_session.scheme,Session#tcp_session.socket),
    httpc_manager:abort_session(Req#request.address,Session#tcp_session.id,
			       aborted_request),
    stop.


error(#request{from=From,ref=Ref,id=Id},Reason) ->
    gen_server:cast(From,{Ref,Id,{error,Reason}}).


