%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.1, (the "License"); you may not use this file except in
%% compliance with the License. You should have received a copy of the
%% Erlang Public License along with this software. If not, it can be
%% retrieved via the world wide web at http://www.erlang.org/.
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Initial Developer of the Original Code is Mobile Arts AB
%% Portions created by Mobile Arts are Copyright 2002, Mobile Arts AB
%% All Rights Reserved.''
%% 
%%
-module(httpc_handler).

-include("http.hrl").

-export([init_connection/2, next_response_with_request/2]).

%%% ==========================================================================
%%% "Main" function in the spawned process for the session.
init_connection(Req,Session) when record(Req,request) ->
    case catch http_lib:connect(Req) of
	{ok,Socket} ->
	    case catch http_request:send(Req,Socket) of
		ok ->
		    case Session#tcp_session.clientclose of
			true ->
			    ok;
			false ->
			    httpc_manager:register_socket(Req#request.address,
							  Session#tcp_session.id,
							  Socket)
		    end,
		    next_response_with_request(Req,
					       Session#tcp_session{socket=Socket});
		{error,Reason} -> % Not possible to use new session
		    http_response:error(Req,Reason),
		    exit_session_ok(Req#request.address,
				    Session#tcp_session{socket=Socket})
	    end;
	{error,Reason} -> % Not possible to set up new session
	    http_response:error(Req,Reason),
	    exit_session_ok2(Req#request.address,
			     Session#tcp_session.clientclose,
			     Session#tcp_session.id)
    end.

next_response_with_request(Req,Session) ->
    Timeout=(Req#request.settings)#client_settings.timeout,
    case catch read(Timeout,Session#tcp_session.scheme,
		    Session#tcp_session.socket) of
	{Status,Headers,Body} ->
	    NewReq=handle_response({Status,Headers,Body},Timeout,Req,Session),
	    next_response_with_request(NewReq,Session);
	{error,Reason} ->
	    http_response:error(Req,Reason),
	    exit_session(Req#request.address,Session,aborted_request);
	{'EXIT',Reason} ->
	    http_response:error(Req,Reason),
	    exit_session(Req#request.address,Session,aborted_request)
    end.

handle_response(Response, Timeout, Request, Session) ->
    case http_response:respond(Response, Request, Session) of
	ok ->
	    next_response(Timeout,Request#request.address,Session);
	stop ->
	    exit(normal);
	{error,Reason} ->
	    http_response:error(Request,Reason),
	    exit_session(Request#request.address,Session,aborted_request)
    end.



%%% Wait for the next respond until
%%% - session is closed by the other side
%%%      => set up a new a session, if there are pending requests in the que
%%% - "Connection:close" header is received
%%%      => close the connection (release socket) then
%%%         set up a new a session, if there are pending requests in the que
%%% 
%%% Note:
%%% - When invoked there are no pending responses on received requests.
%%% - Never close the session explicitly, let it timeout instead!
next_response(Timeout,Address,Session) ->
    case httpc_manager:next_request(Address,Session#tcp_session.id) of
	no_more_requests ->
	    %% There are no more pending responses, now just wait for
	    %% timeout or a new response.
	    case catch read(Timeout,
			    Session#tcp_session.scheme,
			    Session#tcp_session.socket) of
		{error,Reason} when Reason==session_remotely_closed;
				    Reason==session_local_timeout ->
		    exit_session_ok(Address,Session);
		{error,Reason} ->
		    exit_session(Address,Session,aborted_request);
		{'EXIT',Reason} ->
		    exit_session(Address,Session,aborted_request);
		{Status2,Headers2,Body2} ->
		    case httpc_manager:next_request(Address,
						    Session#tcp_session.id) of
			no_more_requests -> % Should not happen!
			    exit_session(Address,Session,aborted_request);
			{error,Reason} -> % Should not happen!
			    exit_session(Address,Session,aborted_request);
			NewReq ->
			    handle_response({Status2,Headers2,Body2},
					    Timeout,NewReq,Session)
		    end
	    end;
	{error,Reason} -> % The connection has been closed by httpc_manager
	    exit_session(Address,Session,aborted_request);
	NewReq ->
	    NewReq
    end.

%% ===========================================================================
%% Internals

%%% Read in and parse response data from the socket
read(Timeout,SockType,Socket) ->
    Info=#response{scheme=SockType,socket=Socket},
    
    http_lib:setopts(SockType,Socket,[binary,{packet, 0}]),
    Msg = http_lib:receive_packet(SockType, Socket, 0, Timeout),
    {StatusCode, Headers, Body} = http_response:parse_message(Msg),
    
    case Headers#http_response_h.content_type of
	"multipart/byteranges" ++ Param ->
	    exit(not_yet_implemented);
	_ ->
	    {NewHeaders, NewBody} = http_response:parse_body(Headers, Body),
	    {StatusCode, NewHeaders, NewBody}
    end.
    

exit_session(Address,Session,Reason) ->
    http_lib:close(Session#tcp_session.scheme,Session#tcp_session.socket),
    httpc_manager:abort_session(Address,Session#tcp_session.id,Reason),
    exit(normal).

%%% This is the "normal" case to close a persistent connection. I.e., there are
%%% no more requests waiting and the session was closed by the client, or
%%% server because of a timeout or user request.
exit_session_ok(Address,Session) ->
    http_lib:close(Session#tcp_session.scheme,Session#tcp_session.socket),
    exit_session_ok2(Address,Session#tcp_session.clientclose,
		     Session#tcp_session.id).

exit_session_ok2(Address,ClientClose,Sid) ->
    case ClientClose of
	false ->
	    httpc_manager:close_session(Address,Sid);
	true ->
	    ok
    end,
    exit(normal).
